@@ -11,6 +11,7 @@
     [clojure.java.io :as jio]
     [clojure.pprint :refer [pprint]]
     [clojure.string :as str]
+    [clojure.tools.deps.alpha.util.concurrent :as concurrent]
     [clojure.tools.deps.alpha.util.dir :as dir]
     [clojure.tools.deps.alpha.extensions :as ext]
 
@@ -24,6 +25,8 @@
     [clojure.lang PersistentQueue]
     [java.io File]))
 
+(set! *warn-on-reflection* true)
+
 (def ^:private merge-alias-rules
   {:deps merge
    :extra-deps merge
@@ -165,54 +168,90 @@
   (when trace?
     (conj trace (merge entry {:include include :reason reason}))))
 
+(defn- next-path
+  [pendq q on-error]
+  (let [[fchild & rchildren] pendq]
+    (if fchild
+      {:path fchild, :pendq rchildren, :q' q}
+      (let [next-q (peek q)
+            q' (pop q)]
+        (if (map? next-q)
+          (let [{:keys [pend-children ppath]} next-q
+                result @pend-children]
+            (when (instance? Throwable result)
+              (on-error result))
+            (next-path (map #(conj ppath %) result) q' on-error))
+          {:path next-q, :q' q'})))))
+
 (defn- expand-deps
-  [deps default-deps override-deps config trace?]
-  (loop [q (into (PersistentQueue/EMPTY) (map vector deps))
+  [deps default-deps override-deps config executor trace?]
+  (loop [pendq nil
+         q (into (PersistentQueue/EMPTY) (map vector deps))
          version-map nil
          exclusions nil
-         trace []] ;; path to set of exclusions at path
-    (if-let [path (peek q)] ;; path from root dep to dep being expanded
-      (let [q' (pop q)
-            [lib coord] (peek path)
-            parents (pop path)
-            override-coord (get override-deps lib)
-            use-coord (cond override-coord override-coord
-                            coord coord
-                            :else (get default-deps lib))
-            coord-id (ext/dep-id lib use-coord config)
-            entry (cond-> {:path parents, :lib lib, :coord coord, :use-coord use-coord, :coord-id coord-id}
-                    override-coord (assoc :override-coord override-coord))
-            {:keys [include reason]} (include-coord? version-map lib parents exclusions)]
-        (if include
-          (let [use-path (conj parents lib)
-                {:deps/keys [manifest root] :as manifest-info} (ext/manifest-type lib use-coord config)
-                use-coord (merge use-coord manifest-info)
-                children (dir/with-dir (if root (jio/file root) dir/*the-dir*)
-                           (canonicalize-deps (ext/coord-deps lib use-coord manifest config) config))
-                child-paths (map #(conj use-path %) children)
-                {:keys [include reason vmap]} (add-coord version-map lib coord-id use-coord parents reason config)]
-            (if include
-              (let [excl' (if-let [excl (:exclusions use-coord)]
-                            (add-exclusion exclusions use-path excl)
-                            exclusions)]
-                (recur (into q' child-paths) vmap excl' (trace+ trace? trace entry include reason)))
-              (recur q' vmap exclusions (trace+ trace? trace entry include reason))))
-          (recur q' version-map exclusions (trace+ trace? trace entry include reason))))
-      (cond-> version-map trace? (with-meta {:trace {:log trace, :vmap version-map, :exclusions exclusions}})))))
+         trace []]
+    (let [{:keys [path pendq q']} (next-path pendq q #(do
+                                                        (concurrent/shutdown-on-error executor)
+                                                        (throw ^Throwable %)))]
+      (if path
+        (let [[lib coord] (peek path)
+              parents (pop path)
+              override-coord (get override-deps lib)
+              use-coord (cond override-coord override-coord
+                              coord coord
+                              :else (get default-deps lib))
+              coord-id (ext/dep-id lib use-coord config)
+              entry (cond-> {:path parents, :lib lib, :coord coord, :use-coord use-coord, :coord-id coord-id}
+                      override-coord (assoc :override-coord override-coord))
+              {:keys [include reason]} (include-coord? version-map lib parents exclusions)]
+          (if include
+            (let [use-path (conj parents lib)
+                  {:deps/keys [manifest root] :as manifest-info} (ext/manifest-type lib use-coord config)
+                  use-coord (merge use-coord manifest-info)
+                  children-fut (dir/with-dir (if root (jio/file root) dir/*the-dir*)
+                             (concurrent/submit-task executor
+                               #(try
+                                  (canonicalize-deps (ext/coord-deps lib use-coord manifest config) config)
+                                  (catch Throwable t t))))
+                  {:keys [include reason vmap]} (add-coord version-map lib coord-id use-coord parents reason config)]
+              (if include
+                (let [excl' (if-let [excl (:exclusions use-coord)]
+                              (add-exclusion exclusions use-path excl)
+                              exclusions)]
+                  (recur pendq (conj q' {:pend-children children-fut, :ppath use-path}) vmap excl' (trace+ trace? trace entry include reason)))
+                (recur pendq q' vmap exclusions (trace+ trace? trace entry include reason))))
+            (recur pendq q' version-map exclusions (trace+ trace? trace entry include reason))))
+        (cond-> version-map trace? (with-meta {:trace {:log trace, :vmap version-map, :exclusions exclusions}}))))))
 
 (defn- lib-paths
-  [version-map config]
+  [version-map]
   (reduce
     (fn [ret [lib {:keys [select versions paths]}]]
       (let [coord (get versions select)
-            paths (->> (get paths select) (map last) (remove nil?) vec)
-            src-paths (ext/coord-paths lib coord (:deps/manifest coord) config)]
-        (assoc ret
-          lib
-          (cond-> (assoc coord :paths src-paths)
-            (seq paths) (assoc :dependents paths)))))
+            paths (->> (get paths select) (map last) (remove nil?) vec)]
+        (assoc ret lib (cond-> coord (seq paths) (assoc :dependents paths)))))
     {} version-map))
 
+(defn- download-libs
+  [executor lib-map config]
+  (let [lib-futs (reduce-kv
+                   (fn [fs lib coord]
+                     (let [fut (concurrent/submit-task
+                                 executor
+                                 #(try
+                                    (ext/coord-paths lib coord (:deps/manifest coord) config)
+                                    (catch Throwable t t)))]
+                       (assoc fs lib fut)))
+                   {} lib-map)]
+    (reduce-kv (fn [lm lib fut]
+                 (let [result @fut]
+                   (if (instance? Throwable result)
+                     (do
+                       (concurrent/shutdown-on-error executor)
+                       (throw ^Throwable result))
+                     (assoc-in lm [lib :paths] result))))
+      lib-map lib-futs)))
+
 (defn resolve-deps
   "Takes a deps configuration map and resolves the transitive dependency graph
   from the initial set of deps. args-map is a map with several keys (all
@@ -224,17 +263,24 @@
 
   settings is an optional map of settings:
     :trace - boolean. If true, the returned lib map will have metadata with :trace log
+    :threads - int. If provided, sets the number of concurrent download threads
 
   Returns a lib map (map of lib to coordinate chosen)."
   ([deps-map args-map]
     (resolve-deps deps-map args-map nil))
   ([deps-map args-map settings]
    (let [{:keys [extra-deps default-deps override-deps]} args-map
+         n (if-let [threads (:threads settings)]
+             (Long/parseLong threads)
+             concurrent/processors)
+         executor (concurrent/new-executor n)
          deps (merge (:deps deps-map) extra-deps)
          version-map (-> deps
                        (canonicalize-deps deps-map)
-                       (expand-deps default-deps override-deps deps-map (:trace settings)))]
-     (with-meta (lib-paths version-map deps-map) (meta version-map)))))
+                       (expand-deps default-deps override-deps deps-map executor (:trace settings)))
+         lib-map (lib-paths version-map)
+         lib-map' (download-libs executor lib-map deps-map)]
+     (with-meta lib-map' (meta version-map)))))
 
 (defn- make-tree
   [lib-map]
@@ -279,27 +325,29 @@
 (comment
   (require '[clojure.tools.deps.alpha.util.maven :as mvn])
 
+  (def ex-svc (concurrent/new-executor 2))
+
   (expand-deps {'org.clojure/clojure {:mvn/version "1.9.0"}}
-    nil nil {:mvn/repos mvn/standard-repos} true)
+    nil nil {:mvn/repos mvn/standard-repos} ex-svc true)
 
   (expand-deps {'org.apache.xmlgraphics/batik-transcoder {:mvn/version "1.7"}}
-               nil nil {:mvn/repos mvn/standard-repos} true)
+               nil nil {:mvn/repos mvn/standard-repos} ex-svc true)
 
   (expand-deps {'org.clojure/clojure {:mvn/version "1.9.0"}
                 'org.clojure/core.memoize {:mvn/version "0.5.8"}}
-    nil nil {:mvn/repos mvn/standard-repos} true)
+    nil nil {:mvn/repos mvn/standard-repos} ex-svc true)
 
   (expand-deps {'org.clojure/clojure {:mvn/version "1.9.0"}
                 'org.clojure/clojurescript {:mvn/version "1.9.946"}
                 'reagent/reagent {:mvn/version "0.6.0"}}
-    nil nil {:mvn/repos mvn/standard-repos} true)
+    nil nil {:mvn/repos mvn/standard-repos} ex-svc true)
 
   (expand-deps {'org.clojure/core.async {:mvn/version "0.3.426"}
                 'reagent/reagent {:mvn/version "0.6.0"}}
-    nil nil {:mvn/repos mvn/standard-repos} true)
+    nil nil {:mvn/repos mvn/standard-repos} ex-svc true)
 
   (expand-deps {'org.clojure/tools.deps.alpha {:mvn/version "0.4.277"}}
-    nil nil {:mvn/repos mvn/standard-repos} true)
+    nil nil {:mvn/repos mvn/standard-repos} ex-svc true)
 
   (print-tree
     (resolve-deps {:deps {'org.clojure/clojure {:mvn/version "1.8.0"}
